
#include "mods.h"





BoundsTree::DeleteTree(Leaf *l)
	{
	Leaf *tlist[4];
	for (int i = 0; i < 4; i++)
		{
		tlist[i] = l->GetQuad(i);
		}
	delete l;
	l = NULL;

	for (i = 0; i < 4; i++)
		{
		if (tlist[i])
			DeleteTree(tlist[i]);
		}
	}

void BoundsTree::LayoutTree(Leaf *l, int depth)
	{
	depth--;
	if (depth == 0) 
		{
		l->SetBottom(TRUE);
		return;
		}

	
	for (int i =0; i < 4; i++)
		{
		l->SetQuad(i, new Leaf(l->GetBounds(i).min,l->GetBounds(i).max));
		}

	LayoutTree(l->GetQuad(0),depth);
	LayoutTree(l->GetQuad(1),depth);
	LayoutTree(l->GetQuad(2),depth);
	LayoutTree(l->GetQuad(3),depth);

	}



BoundsTree::RecurseTree(Leaf *l, int index, Box2D box)

{
	if (l->IsBottom())
		{
		BoundsList temp;
		Point3 norm = fnorms[index];
//fix set a direction here	
		if (invertNormals)
			{
			temp.faceIndex = index;
			l->faceIndex.Append(1,&temp,10);
			}
		else
			{
			if (norm.z > 0.0f)
				{
				temp.faceIndex = index;
				l->faceIndex.Append(1,&temp,10);
				}
			}
		}
	else
		{
		if ((l->GetQuad(0)) && (l->InQuad(0,box))) 
			RecurseTree(l->GetQuad(0),index,box);
		if ((l->GetQuad(1)) && (l->InQuad(1,box))) 
			RecurseTree(l->GetQuad(1),index,box);
		if ((l->GetQuad(2)) && (l->InQuad(2,box))) 
			RecurseTree(l->GetQuad(2),index,box);
		if ((l->GetQuad(3)) && (l->InQuad(3,box))) 
			RecurseTree(l->GetQuad(3),index,box);

		}
}

BoundsTree::AddFace(int index, Box2D box)
	{
	if (head)
		{
		if ((head->GetQuad(0)) && (head->InQuad(0,box))) 
			RecurseTree(head->GetQuad(0),index,box);
		if ((head->GetQuad(1)) && (head->InQuad(1,box))) 
			RecurseTree(head->GetQuad(1),index,box);
		if ((head->GetQuad(2)) && (head->InQuad(2,box))) 
			RecurseTree(head->GetQuad(2),index,box);
		if ((head->GetQuad(3)) && (head->InQuad(3,box))) 
			RecurseTree(head->GetQuad(3),index,box);
		}	
	}


BoundsTree::UpdateTreeFromBaked()
	{
	Matrix3 tm(1);
	BuildTree(&initMsh,tm,tm,NULL,1.0f,FALSE);
	}

BOOL BoundsTree::TrimRecurseTree(Leaf *l)
	{
//if is bottom and empty return TRUE
	if (l->IsBottom() && l->faceIndex.Count() ==0)
		return TRUE;
	else if (l->IsBottom() && l->faceIndex.Count() !=0)
		return FALSE;
	else if ( TrimRecurseTree(l->GetQuad(0)) &&
		 TrimRecurseTree(l->GetQuad(1)) &&
		 TrimRecurseTree(l->GetQuad(2)) &&
		 TrimRecurseTree(l->GetQuad(3)) )
		{
//DebugPrint("Leaf trimmed\n");
		Leaf *deadLeaf1 = l->GetQuad(0);
		Leaf *deadLeaf2 = l->GetQuad(1);
		Leaf *deadLeaf3 = l->GetQuad(2);
		Leaf *deadLeaf4 = l->GetQuad(3);

		delete deadLeaf1;
		delete deadLeaf2;
		delete deadLeaf3;
		delete deadLeaf4;
		l->SetQuad(0, NULL);
		l->SetQuad(1, NULL);
		l->SetQuad(2, NULL);
		l->SetQuad(3, NULL);
		l->SetBottom(TRUE);

		if (l->faceIndex.Count() ==0)
			return TRUE;
		else return FALSE;
		}
	else return FALSE;

	}
void BoundsTree::TrimTree()
	{
	TrimRecurseTree(head);
	}

BoundsTree::BuildTree(Mesh *m, Matrix3 basetm, Matrix3 tm,ViewExp *vpt, float mapSize, BOOL buildZMap)
	{

	if (head) DeleteTree(head);

	msh = *m;


	
	origPoints.SetCount(msh.numVerts);

	int i;
	for (i =0; i < msh.numVerts; i++)
		{
		origPoints[i] = msh.verts[i];
		}

	invTm = Inverse(basetm*tm);


	GraphicsWindow *gw = vpt->getGW();
	
	Matrix3 temptm = basetm * tm;
	gw->setTransform(Matrix3(1));
	Point3 *verts = msh.verts;
	for (i =0; i < msh.numVerts; i++)
		{
		Point3 inPoint,outPoint;
		inPoint = *verts * basetm;
		DWORD flag = gw->transPoint(&inPoint, &outPoint);
		inPoint = inPoint * tm;
		if (buildZMap)
			*verts = outPoint *mapSize;
		else *verts = outPoint ;
		(*verts).z = inPoint.z;
		verts++;
		}

	BuildNormals();

	//layout tree based on depth
	Point2 min;
	min.x = 0.0f;
	min.y = 0.0f;
	Point2 max;
	max.x = 1.0f;
	max.y = 1.0f;

	msh.buildBoundingBox();
	Box3 bb = msh.getBoundingBox();

	min.x = bb.pmin.x;
	min.y = bb.pmin.y;

	max.x = bb.pmax.x;
	max.y = bb.pmax.y;

	head = new Leaf(min,max);
	LayoutTree(head, depth);


//now stuff the points into the boudningbox list
	boundingBoxList.SetCount(msh.numFaces);
	Face *faces = msh.faces;
	for (i =0; i < msh.numFaces; i++)
		{
		boundingBoxList[i].SetEmpty();
		for (int j = 0; j < 3; j++)
			{
			int vid = (*faces).v[j];
			Point3 p = msh.verts[vid];
			Point2 p2;
			p2.x = p.x;
			p2.y = p.y;
			boundingBoxList[i] += p2;

			}
		faces++;
		}
//now go through our tree adding the bounding list 
	for (i =0; i < msh.numFaces; i++)
		{
		AddFace(i,boundingBoxList[i]);
		}
//trim the tree of empty quads
//	TrimTree();
//build ztable
	if (zBuffer)
		delete [] zBuffer;
	zBuffer = NULL;

	if (buildZMap)
		{
		int x = gw->getWinSizeX()*mapSize;
		int y = gw->getWinSizeY()*mapSize;
		this->w = x;
		this->h = y;
		zSize = this->w*this->h;

		zBuffer = new ZTable[x*y];
		ZTable *zt = zBuffer;
		float sx = 0.0f;
		float sy = 0.0f;
	
		for ( i = 0; i < y; i++)
			{
			for (int j = 0; j < x; j++)
				{
				Point3 hitPoint;
				hitPoint.x = sx;
				hitPoint.y = sy;
				hitPoint.z = 0.0f;
				zt->faceIndex = -1;
				float z;
				Point3 norm, bary;
				DWORD findex;
				if (HitMesh(hitPoint, z, findex, norm, bary))
					{
					zt->faceIndex =findex;
					Point3 a,b,c;
					int ia,ib,ic;
					ia = msh.faces[findex].v[0];
					ib = msh.faces[findex].v[1];
					ic = msh.faces[findex].v[2];
					a = origPoints[ia];
					b = origPoints[ib];
					c = origPoints[ic];
					zt->p = a * bary.x +  b * bary.y +  c * bary.z;
					}
				sx += 1.0f;
				zt++;
				}
			sy += 1.0f;
			sx = 0.0f;
			}	
	
		}

	}

void BoundsTree::BuildNormals()
{

	Mesh *mesh;
	mesh = &msh;
	Face *face;	
	Point3 *verts;
	Point3 v0, v1, v2;


	face = mesh->faces;	
	verts = mesh->verts;
	vnorms.SetCount(mesh->getNumVerts());
	fnorms.SetCount(mesh->getNumFaces());

	// Compute face and vertex surface normals
	for (int i = 0; i < mesh->getNumVerts(); i++) 
		{
		vnorms[i] = VNormal();
		}
	for (i = 0; i < mesh->getNumFaces(); i++, face++) 
		{

// Calculate the surface normal
		v0 = verts[face->v[0]];
		v1 = verts[face->v[1]];
		v2 = verts[face->v[2]];
		fnorms[i] = (v1-v0)^(v2-v1);
		for (int j=0; j<3; j++) 
			{		
			vnorms[face->v[j]].AddNormal(fnorms[i],face->smGroup);
			}
		fnorms[i] = Normalize(fnorms[i]);
		}
	for (i=0; i < mesh->getNumVerts(); i++) 
		{
		vnorms[i].Normalize();
		}
//FIX move to ornormal loop
	
}

#define EPSILON 0.000001f


BOOL BoundsTree::IsInsideMesh(Point3 hitPoint)
{

Leaf *l = head;
int ct = 0;
BOOL hit = FALSE;
if (l == NULL) 
	{
	return FALSE;
	}

while ( (l!=NULL) && (l->IsBottom() == FALSE))
	{
	int id = l->InWhichQuad(hitPoint);
	l = l->GetQuad(id);
	ct++;
	}

ct = 0;

if (l)
	{
	if (l->faceIndex.Count() == 0) 
		return FALSE;
	else
		{
		for (int i = 0; i < l->faceIndex.Count(); i++)
			{
			int faceIndex = l->faceIndex[i].faceIndex;
			Box2D b = boundingBoxList[faceIndex];
			if ( (hitPoint.x >= b.min.x) && (hitPoint.x <= b.max.x) &&
				 (hitPoint.y >= b.min.y) && (hitPoint.y <= b.max.y) )
				{
//now check bary coords
				Point3 a,b,c;
				a = msh.verts[msh.faces[faceIndex].v[0]];
				b = msh.verts[msh.faces[faceIndex].v[1]];
				c = msh.verts[msh.faces[faceIndex].v[2]];
				Point3 az,bz,cz,hitPointZ;
				az = a;
				bz = b;
				cz = c;
				az.z = 0.0f;
				bz.z = 0.0f;
				cz.z = 0.0f;
				hitPointZ = hitPoint;
				hitPointZ.z = 0.0f;
	
				Point3 bry;
				bry = BaryCoords(az, bz, cz, hitPointZ);
//if inside bary find the the z point			
				if (!( (bry.x<0.0f || bry.x>1.0f || bry.y<0.0f || bry.y>1.0f || bry.z<0.0f || bry.z>1.0f) ||
					(fabs(bry.x + bry.y + bry.z - 1.0f) > EPSILON) ))
					{
					float tz = a.z * bry.x + b.z * bry.y + c.z * bry.z; 
					if  (tz < hitPoint.z )  ct++;
					
					}
				}
			}
		}	
	}
if ((ct%2) == 1) return TRUE;
else return FALSE;

}

BOOL BoundsTree::HitZBuffer(IPoint2 m, DWORD &findex, Point3 &p)
{
	if (zBuffer)
		{
		int pos;
		pos = m.y * w + m.x;
		if ( (pos < zSize) && (pos >= 0) )
			{
			if (zBuffer[pos].faceIndex != -1)
				{
				findex = zBuffer[pos].faceIndex;
				p = zBuffer[pos].p;
				
				return TRUE;
				}
			}
		}
	return FALSE;
}

BOOL BoundsTree::HitQuadTree(IPoint2 m, DWORD &findex, Point3 &p, Point3 &bary, float &finalZ)
{
Leaf *l = head;
BOOL hit = FALSE;
Point3 hitPoint(0.0f,0.0f,0.0f);

hitPoint.x = (float) m.x;
hitPoint.y = (float) m.y;
float z = 0.0f;
Point3 bry;
if (l == NULL) 
	{
	z = 0.0f;
	return FALSE;
	}
int ct = 0;
while ( (l!=NULL) && (l->IsBottom() == FALSE))
	{
	int id = l->InWhichQuad(hitPoint);
	l = l->GetQuad(id);
	ct++;
	}
if (l)
	{
	if (l->faceIndex.Count() == 0) 
		return FALSE;
	else
		{
		for (int i = 0; i < l->faceIndex.Count(); i++)
			{
			int faceIndex = l->faceIndex[i].faceIndex;
			Box2D b = boundingBoxList[faceIndex];
			if ( (hitPoint.x >= b.min.x) && (hitPoint.x <= b.max.x) &&
				 (hitPoint.y >= b.min.y) && (hitPoint.y <= b.max.y) )
				{
//now check bary coords
				Point3 a,b,c;
				a = msh.verts[msh.faces[faceIndex].v[0]];
				b = msh.verts[msh.faces[faceIndex].v[1]];
				c = msh.verts[msh.faces[faceIndex].v[2]];
				Point3 az,bz,cz,hitPointZ;
				az = a;
				bz = b;
				cz = c;
				az.z = 0.0f;
				bz.z = 0.0f;
				cz.z = 0.0f;
				hitPointZ = hitPoint;
				hitPointZ.z = 0.0f;
	
				Point3 bry;
				bry = BaryCoords(az, bz, cz, hitPointZ);
//if inside bary find the the z point			
				if (!( (bry.x<0.0f || bry.x>1.0f || bry.y<0.0f || bry.y>1.0f || bry.z<0.0f || bry.z>1.0f) ||
					(fabs(bry.x + bry.y + bry.z - 1.0f) > EPSILON) ))
					{
					float tz = a.z * bry.x + b.z * bry.y + c.z * bry.z; 
					if ( (tz > z ) || (hit == FALSE) )  
						{
						z = tz;
						findex = faceIndex;
						bary = bry;
						finalZ = z;
						}
					hit = TRUE;
					}
				}
			}
		}	
	}
if (hit)
	{
	Point3 a,b,c;
	int ia,ib,ic;
	ia = msh.faces[findex].v[0];
	ib = msh.faces[findex].v[1];
	ic = msh.faces[findex].v[2];
	a = origPoints[ia];
	b = origPoints[ib];
	c = origPoints[ic];
	p = a * bary.x +  b * bary.y +  c * bary.z;
	}

return hit;

}


BOOL BoundsTree::HitMesh(Point3 hitPoint, float &z, DWORD &findex, Point3 &norm, Point3 &bary)
{

Leaf *l = head;
BOOL hit = FALSE;
if (l == NULL) 
	{
	z = 0.0f;
	return FALSE;
	}
int ct = 0;
while ( (l!=NULL) && (l->IsBottom() == FALSE))
	{
	int id = l->InWhichQuad(hitPoint);
	l = l->GetQuad(id);
	ct++;
	}
if (l)
	{
	if (l->faceIndex.Count() == 0) 
		return FALSE;
	else
		{
		for (int i = 0; i < l->faceIndex.Count(); i++)
			{
			int faceIndex = l->faceIndex[i].faceIndex;
			Box2D b = boundingBoxList[faceIndex];
			if ( (hitPoint.x >= b.min.x) && (hitPoint.x <= b.max.x) &&
				 (hitPoint.y >= b.min.y) && (hitPoint.y <= b.max.y) )
				{
//now check bary coords
				Point3 a,b,c;
				a = msh.verts[msh.faces[faceIndex].v[0]];
				b = msh.verts[msh.faces[faceIndex].v[1]];
				c = msh.verts[msh.faces[faceIndex].v[2]];
				Point3 az,bz,cz,hitPointZ;
				az = a;
				bz = b;
				cz = c;
				az.z = 0.0f;
				bz.z = 0.0f;
				cz.z = 0.0f;
				hitPointZ = hitPoint;
				hitPointZ.z = 0.0f;
	
				Point3 bry;
				bry = BaryCoords(az, bz, cz, hitPointZ);
//if inside bary find the the z point			
				if (!( (bry.x<0.0f || bry.x>1.0f || bry.y<0.0f || bry.y>1.0f || bry.z<0.0f || bry.z>1.0f) ||
					(fabs(bry.x + bry.y + bry.z - 1.0f) > EPSILON) ))
					{
					float tz = a.z * bry.x + b.z * bry.y + c.z * bry.z; 
					if ( (tz > z ) || (hit == FALSE) )  
						{
						z = tz;
						findex = faceIndex;
						norm = fnorms[faceIndex];
						bary = bry;
						}
					hit = TRUE;
					}
				}
			}
		}	
	}

return hit;
}




BOOL BoundsTree::HitMesh(Point3 hitPoint, Point3 &norm)
{
norm = Point3 (0.0f,0.0f,0.0f);
Leaf *l = head;
BOOL hit = FALSE;
int nia,nib,nic;
Point3 ibary;
DWORD smgroup;
int ifaceID;

float z = 0.0f;

if (l == NULL) 
	{
	
	return FALSE;
	}
int ct = 0;
while ( (l!=NULL) && (l->IsBottom() == FALSE))
	{
	int id = l->InWhichQuad(hitPoint);
	l = l->GetQuad(id);
	ct++;
	}
if (l)
	{
	if (l->faceIndex.Count() == 0) 
		return FALSE;
	else
		{
		for (int i = 0; i < l->faceIndex.Count(); i++)
			{
			int faceIndex = l->faceIndex[i].faceIndex;
			Box2D b = boundingBoxList[faceIndex];
			if ( (hitPoint.x >= b.min.x) && (hitPoint.x <= b.max.x) &&
				 (hitPoint.y >= b.min.y) && (hitPoint.y <= b.max.y) )
				{
//now check bary coords
				int ia,ib,ic;
				
				ia = msh.faces[faceIndex].v[0];
				ib = msh.faces[faceIndex].v[1];
				ic = msh.faces[faceIndex].v[2];
				Point3 a,b,c;
				a = msh.verts[ia];
				b = msh.verts[ib];
				c = msh.verts[ic];
				Point3 az,bz,cz,hitPointZ;
				az = a;
				bz = b;
				cz = c;
				az.z = 0.0f;
				bz.z = 0.0f;
				cz.z = 0.0f;
				hitPointZ = hitPoint;
				hitPointZ.z = 0.0f;
	
				Point3 bry;
				bry = BaryCoords(az, bz, cz, hitPointZ);
//if inside bary find the the z point			
				if (!( (bry.x<0.0f || bry.x>1.0f || bry.y<0.0f || bry.y>1.0f || bry.z<0.0f || bry.z>1.0f) ||
					(fabs(bry.x + bry.y + bry.z - 1.0f) > EPSILON) ))
					{
					float tz = a.z * bry.x + b.z * bry.y + c.z * bry.z; 
					if ( (tz > z ) || (hit == FALSE) )  
						{
						z = tz;
						hit = TRUE;
						nia = ia;
						nib = ib;
						nic = ic;
						ifaceID = faceIndex;
						smgroup = msh.faces[faceIndex].getSmGroup();
						ibary = bry;
						}
					}
				}
			}
		}	
	}
if (hit)
	{
	Point3 normA, normB, normC;
	normA = vnorms[nia].GetNormal(smgroup);
	normB = vnorms[nib].GetNormal(smgroup);
	normC = vnorms[nic].GetNormal(smgroup);
	norm = (normA * ibary.x) + (normB * ibary.y) + (normC * ibary.z);
//	norm = fnorms[ifaceID];
	}

return hit;
}


BoundsTree::SetInitMesh(Mesh *m)
	{
	initMsh = *m;
	}
Mesh BoundsTree::GetInitMesh()
	{
	return initMsh;
	}

IOResult BoundsTree::Load(ILoad *iload)
	{
	ULONG nb;
	IOResult res;
	int id;
	int fct = 0;
	int vct = 0;
	while (IO_OK==(res=iload->OpenChunk())) {
		switch(id=iload->CurChunkID())  {
			case MESH_VERTEXCOUNT_CHUNK:
				iload->Read(&vct,sizeof(vct),&nb);
				initMsh.setNumVerts(vct);
				break;
			case MESH_FACECOUNT_CHUNK:
				iload->Read(&fct,sizeof(fct),&nb);
				initMsh.setNumFaces(fct);
				break;
			case MESH_VERTEXDATA_CHUNK:
				iload->Read(initMsh.verts,sizeof(Point3)*vct,&nb);
				
				break;
			case MESH_FACEDATA_CHUNK:
				iload->Read(initMsh.faces,sizeof(Face)*fct,&nb);
				break;
			}
		iload->CloseChunk();
		if (res!=IO_OK) 
			return res;
		}


	return IO_OK;
	
	}

IOResult BoundsTree::Save(ISave *isave) 
	{
	ULONG nb;
	// Save common stuff

	int vct,fct;
	vct = initMsh.numVerts;
	fct = initMsh.numFaces;
	//save vertex count
	isave->BeginChunk(MESH_VERTEXCOUNT_CHUNK);
	isave->Write(&vct,sizeof(vct),&nb);
	isave->EndChunk();

	//save face count
	isave->BeginChunk(MESH_FACECOUNT_CHUNK);
	isave->Write(&fct,sizeof(fct),&nb);
	isave->EndChunk();

	//save vertex data
	isave->BeginChunk(MESH_VERTEXDATA_CHUNK);
	isave->Write(initMsh.verts,sizeof(Point3)*vct,&nb);
	isave->EndChunk();

	//save face data
	isave->BeginChunk(MESH_FACEDATA_CHUNK);
	isave->Write(initMsh.faces,sizeof(Face)*fct,&nb);
	isave->EndChunk();

	return IO_OK;

	}


void BoundsTree::BuildVertLoops()
{
//build adjacent list
	AdjEdgeList edges(msh);
//loop through verts
	for (int i = 0; i <vertLoops.Count(); i++)
		if (vertLoops[i]) delete vertLoops[i];
	vertLoops.SetCount(msh.numVerts);
	for (i = 0; i < msh.numVerts; i++)
		vertLoops[i] = new LoopData();
//loop through edges see if on is open if so start from there
	for (int j = 0; j < edges.edges.Count();j++)
		{
		if (edges.edges[j].Visible(msh.faces))
			{
			if (edges.edges[j].v[0]>=0)
				{
				int id = edges.edges[j].v[1];
				vertLoops[edges.edges[j].v[0]]->verts.Append(1,&id);
				}
			if (edges.edges[j].v[1]>=0)
				{
				int id = edges.edges[j].v[0];
				vertLoops[edges.edges[j].v[1]]->verts.Append(1,&id);
				}
			}
		}
	BuildFaceConnection();
}


Point3 BoundsTree::SmoothVert(int index, BOOL addative, Tab<Point3> *deltaList )
{
int ct = vertLoops[index]->verts.Count();
Point3 center(0.0f,0.0f,0.0f);
Point3 p;
for (int i =0; i < ct; i++)
	{
	int id = vertLoops[index]->verts[i];

	if (addative)
		p = origPoints[id]+(*deltaList)[id] ;
	else p = origPoints[id] ;
//DebugPrint("Id %d %f %f %f\n",id,p.x,p.y,p.z);

	center += p;
	}
center = (center/(float)ct);

Point3 op = origPoints[index];
//DebugPrint("op %f %f %f new Point %f %f %f\n",op.x,op.y,op.z,center.x,center.y,center.z);
return  center - op;
}


void BoundsTree::BuildFaceConnection()
{
//loop through verts
	for (int i = 0; i <faceConnectionList.Count(); i++)
		if (faceConnectionList[i]) delete faceConnectionList[i];
	faceConnectionList.SetCount(msh.numVerts);
	for (i = 0; i < msh.numVerts; i++)
		faceConnectionList[i] = new FaceConnectionData();
//loop through edges see if on is open if so start from there
	for (int j = 0; j < msh.numFaces;j++)
		{
		int a = msh.faces[j].v[0];
		faceConnectionList[a]->faces.Append(1,&j);
		a = msh.faces[j].v[1];
		faceConnectionList[a]->faces.Append(1,&j);
		a = msh.faces[j].v[2];
		faceConnectionList[a]->faces.Append(1,&j);
		}
}



void BoundsTree::SelectFace(int index, BitArray &f, BOOL set)
{
if (f.GetSize() < msh.numFaces)
	f.SetSize(msh.numFaces);

for (int i = 0; i < faceConnectionList[index]->faces.Count(); i++)
	{
	int a = faceConnectionList[index]->faces[i];
	f.Set(a,set);
	}

/*
for (int i = 0; i < msh.numFaces; i++)
	{
	
	if (index == msh.faces[i].v[0])
		f.Set(i,set);
	else if (index == msh.faces[i].v[1])
		f.Set(i,set);
	else if (index == msh.faces[i].v[2])
		f.Set(i,set);

	}
*/
}