
#define MESH_CHUNK 0x1010

#define MESH_VERTEXCOUNT_CHUNK 0x1020
#define MESH_FACECOUNT_CHUNK 0x1030

#define MESH_VERTEXDATA_CHUNK 0x1040
#define MESH_FACEDATA_CHUNK 0x1050


typedef struct 
	{ 
	Point3 p; //point in local space
	int faceIndex; 
	} ZTable;

class VNormal {
	public:
		Point3 norm;
		DWORD smooth;
		VNormal *next;
		BOOL init;



		VNormal() {smooth=0;next=NULL;init=FALSE;norm=Point3(0,0,0);}
		VNormal(Point3 &n,DWORD s) {next=NULL;init=TRUE;norm=n;smooth=s;}
		~VNormal() {delete next;}
		void AddNormal(Point3 &n,DWORD s);
		Point3 &GetNormal(DWORD s);
		void Normalize();
	};




class Leaf;

class BoundsList
{
public:
	int faceIndex;
	BoundsList()
		{
		faceIndex = -1;
		}
};

class Leaf
{

private:
	Box2D bounds[4];
	BOOL bottom;
	
public:

	Leaf *quad[4];
	Tab<BoundsList> faceIndex;

	Leaf(Point2 min, Point2 max)
		{
		bottom = FALSE;
		quad[0] = NULL;
		quad[1] = NULL;
		quad[2] = NULL;
		quad[3] = NULL;
		Point2 mid = (min+max)*0.5f;

		bounds[0].min = min;
		bounds[0].max = mid;

		bounds[1].min.x = mid.x;
		bounds[1].min.y = min.y;
		bounds[1].max.x = max.x;
		bounds[1].max.y = mid.y;

		bounds[2].min.x = min.x;
		bounds[2].min.y = mid.y;
		bounds[2].max.x = mid.x;
		bounds[2].max.y = max.y;

		bounds[3].min = mid;
		bounds[3].max = max;

		}
	BOOL InQuad(int i, Box2D box)
		{
		Box2D b = bounds[i];
		if (box.max.x < b.min.x) return FALSE;
		if (box.min.x > b.max.x) return FALSE;

		if (box.max.y < b.min.y) return FALSE;
		if (box.min.y > b.max.y) return FALSE;

		return TRUE;

		}
	int InWhichQuad(Point3 p)
		{
		if ( (p.x >=bounds[0].min.x) && (p.x < bounds[0].max.x) &&
			 (p.y >=bounds[0].min.y) && (p.y < bounds[0].max.y) )
			 return 0;
		else if ( (p.x >=bounds[1].min.x) && (p.x < bounds[1].max.x) &&
			 (p.y >=bounds[1].min.y) && (p.y < bounds[1].max.y) )
			 return 1;
		else if ( (p.x >=bounds[2].min.x) && (p.x < bounds[2].max.x) &&
			 (p.y >=bounds[2].min.y) && (p.y < bounds[2].max.y) )
			 return 2;
		else if ( (p.x >=bounds[3].min.x) && (p.x < bounds[3].max.x) &&
			 (p.y >=bounds[3].min.y) && (p.y < bounds[3].max.y) )
			 return 3;
		return -1;
		}

	Box2D GetBounds(int i) {	if ( (i>=0) && (i<4) ) 
								return bounds[i];
								else return Box2D();}
	
	Leaf *GetQuad(int i) {	if ( (i>=0) && (i<4) ) 
							return quad[i];
							else return NULL;
						}
	void SetQuad(int i, Leaf *l) { if ( (i>=0) && (i<4) ) quad[i] = l;}

	BOOL IsBottom() {return bottom;}
	void SetBottom (BOOL b) { bottom = b;}

};


class LoopData
{
public:
	BOOL closed;
	Tab<int> verts;
};

class FaceConnectionData
{
public:
	BOOL closed;
	Tab<int> faces;
};


class BoundsTree
{


	BOOL invertNormals;
	BOOL normalizeMesh;
	int depth;

	float uoffset, voffset, uscale, vscale;


	

	Tab<VNormal> vnorms;
	Tab<Point3> fnorms;
	

	

	Mesh msh;
	Mesh initMsh;

	void LayoutTree(Leaf *l, int depth);

	RecurseTree(Leaf *l, int index, Box2D box);
	AddFace(int index, Box2D box);
	Tab<Box2D> boundingBoxList;

	void BuildNormals();


public:

	Tab<Point3> origPoints;

	Leaf *head;

	Matrix3 invTm;

	int w,h;
	int zSize;
	ZTable *zBuffer;

	void FreeCaches()
		{
		origPoints.ZeroCount();
		origPoints.Resize(0);

		if (head) DeleteTree(head);
		head = NULL;
		if (zBuffer) delete [] zBuffer;
		for (int i = 0; i <vertLoops.Count(); i++)
			if (vertLoops[i]) delete vertLoops[i];
		vertLoops.ZeroCount();
		for (i = 0; i <faceConnectionList.Count(); i++)
			if (faceConnectionList[i]) delete faceConnectionList[i];
		faceConnectionList.ZeroCount();
		}

	BoundsTree()
		{
		normalizeMesh = TRUE;
		uoffset = 0.0f;
		voffset = 0.0f;
		uscale = 1.0f;
		vscale = 1.0f;
		depth = 6;
		head = NULL;
		zBuffer = NULL;
		invertNormals = TRUE;
		}

	~BoundsTree()
		{
		FreeCaches();
/*
		if (head) DeleteTree(head);
		if (zBuffer) delete [] zBuffer;
		for (int i = 0; i <vertLoops.Count(); i++)
			if (vertLoops[i]) delete vertLoops[i];
		for (i = 0; i <faceConnectionList.Count(); i++)
			if (faceConnectionList[i]) delete faceConnectionList[i];
*/
		}


	DeleteTree(Leaf *l);
	UpdateTreeFromBaked();
	BuildTree(Mesh *m, Matrix3 basetm, Matrix3 tm, ViewExp *vpt, float mapSize, BOOL buildZMap);

	SetInitMesh(Mesh *m);
	Mesh GetInitMesh();
	
	IOResult Load(ILoad *iload);
	IOResult Save(ISave *isave); 

	void CloneFrom(BoundsTree tree)
		{
		initMsh = tree.initMsh;
		}

	BOOL TrimRecurseTree(Leaf *l);
	void TrimTree();

	BOOL IsInsideMesh(Point3 hitPoint);
//	BOOL HitMesh(Point3 hitPoint, float &z, DWORD &findex, Point3 &norm);
	BOOL HitMesh(Point3 hitPoint, float &z, DWORD &findex, Point3 &norm, Point3 &bary);
	BOOL HitMesh(Point3 hitPoint, Point3 &norm);
	BOOL HitZBuffer(IPoint2 m, DWORD &findex, Point3 &p);
	BOOL HitQuadTree(IPoint2 m, DWORD &findex, Point3 &p, Point3 &bary, float &z);


	void InvertNorms(BOOL n) { invertNormals = n;}

	void SetDepth(int d) { depth = d;}
	void SetNormalizeMesh(int n) { normalizeMesh = n;}
	void SetPlacement(float ou, float ov, float su, float sv) 
			{ 
			uoffset = ou;
			voffset = ov;
			uscale = su;
			vscale = sv;
			}

	BOOL buildVertLoops;
	Tab<LoopData*> vertLoops;

	void BuildVertLoops();
	Point3 SmoothVert(int index, BOOL addative, Tab<Point3> *deltaList );

	void BuildFaceConnection();
	Tab<FaceConnectionData*> faceConnectionList;

	void SelectFace(int index, BitArray &fs,BOOL set);
private:
};
