/**********************************************************************
 *<
	FILE: mods.h

	DESCRIPTION:

	CREATED BY: Rolf Berteig (based on prim.h)

	HISTORY:

 *>	Copyright (c) 1994, All Rights Reserved.
 **********************************************************************/

#ifndef __MODS__H
#define __MODS__H

#include "Max.h"
//#include "reslib.h"
#include "modsres.h"
#include "iparamm2.h"
#include "modstack.h"
#include "BoundsTree.h"
#include "icurvctl.h"
#include "meshadj.h"
#include "macrorec.h"
#include "SURF_API.h"
#include "polyobj.h"
#include "include\ipaintmod.h"
#include "include\iink.h"

#define PBLOCK_REF	0



#define USETABLET

#ifdef USETABLET
#include "msgpack.h"
#include <wintab.h>
#define PACKETDATA	( PK_BUTTONS | PK_NORMAL_PRESSURE)
#define PACKETMODE	PK_BUTTONS
#include <pktdef.h>
#endif


TCHAR *GetString(int id);

#define EDITMESH_CLASS_ID			0x00050
#define EDITSPLINE_CLASS_ID			0x00060
#define EDITPATCH_CLASS_ID			0x00070
#define EDITLOFT_CLASS_ID			0x00080

#define CLUSTOSM_CLASS_ID			0x25215824

#define RESET_XFORM_CLASS_ID		0x8d562b81
#define CLUSTNODEOSM_CLASS_ID		0xc4d33

extern ClassDesc* GetPaintModDesc();
// in mods.cpp
extern HINSTANCE hInstance;

// For 'Supports Object of Type' rollups
extern BOOL CALLBACK DefaultSOTProc(HWND hWnd,UINT msg,WPARAM wParam,LPARAM lParam);

#define BIGFLOAT	float(999999)

#define NEWSWMCAT	_T("Modifiers")


#define STARTSTROKE		1
#define INSTROKE		2
#define ENDSTROKE		3
#define OUTOFSTROKE		4


#define HITTEST_RAYCAST		0
#define HITTEST_QUADTREE	1
#define HITTEST_ZBUFFER		2

#define INK_NORMAL			0
#define INK_ORIGINALNORMAL	1
#define INK_REVERT			2
#define INK_SMOOTH			3
#define INK_ZLOCAL			4
#define INK_XYLOCAL			5
#define INK_ZSCREEN			6
#define INK_XYSCREEN		7
#define INK_SELECTVERTS		8
#define INK_SELECTFACES		9

#define NUMBER_DEFAULTINKS	10


#define POINTS_CHUNK	0x0010
#define INK_TYPE_CHUNK	0x0020
#define SELECT_VERTS_CHUNK	0x0030
#define SELECT_FACES_CHUNK	0x0040
#define SELECT_VERTSFLOAT_CHUNK	0x0050


#define LPOINTS_CHUNK	0x0060
#define LSELECT_VERTSFLOAT_CHUNK	0x0070
#define LSELECT_FACES_CHUNK	0x0080

#define PAINTMESH_OBJECT	0
#define PAINTPOLY_OBJECT	1
#define PAINTPATCH_OBJECT	2
#define PAINTNURBS_OBJECT	3



//--- Parameter map/block descriptors -------------------------------



class NullView: public View {
	public:
		Point2 ViewToScreen(Point3 p) { return Point2(p.x,p.y); }
		NullView() { worldToView.IdentityMatrix(); screenW=640.0f; screenH = 480.0f; }
	};


class PaintMode;

class PaintModData : public LocalModData
	{
public:

	int objectType;

	Point3 paintPoint, paintNorm;

	Point3 alignDir;


	BOOL setPoints;

	BOOL draw;
	BOOL isMesh;
	BOOL redoNormalList;
	void BuildNormals(Mesh *mesh);
	Mesh msh;
	Matrix3 backTm;

	Matrix3 tmToWorld;
	Matrix3 tmToView;
	void BuildONormals(Mesh *mesh);
	BOOL rebuildOriginalNorms;
	
	Tab<VNormal> vnorms;
	Tab<Point3> fnorms;
	Tab<Point3> origFNorms;
	Tab<VNormal> origVNorms;
	Tab<Point3> oPointList;
	Tab<float> smoothAmount;


	Tab<Point3> DeltaList;
	Tab<Point3> TempList;
	Tab<float> SelectVertList;
	BitArray selectedFaceList;


	BoundsTree meshSearchTree;

	void FreeTempData()
		{
		vnorms.ZeroCount();
		vnorms.Resize(0);

		fnorms.ZeroCount();
		fnorms.Resize(0);

		origFNorms.ZeroCount();
		origFNorms.Resize(0);

		origVNorms.ZeroCount();
		origVNorms.Resize(0);

		oPointList.ZeroCount();
		oPointList.Resize(0);

		smoothAmount.ZeroCount();
		smoothAmount.Resize(0);

		meshSearchTree.FreeCaches();

		}


	PaintModData()
		{
		setPoints = TRUE;
		rebuildOriginalNorms = TRUE;
		isMesh = FALSE;
		redoNormalList = TRUE;
		draw = FALSE;
		}
	~PaintModData()
		{
//clear out data
		}
	 LocalModData *Clone()
		{
		 PaintModData *pmd = new PaintModData();
//clone point list is all we need
		 pmd->DeltaList = DeltaList;
		 pmd->SelectVertList = SelectVertList;
		 pmd->TempList = TempList;
		 pmd->redoNormalList = redoNormalList;
		 pmd->selectedFaceList = selectedFaceList;
		 return pmd;
		}

	IOResult Load(ILoad *iload);
	IOResult Save(ISave *isave);


	};


#define BRUSH_CHUNK	0x0090
#define BRUSH_INK_CHUNK	0x0100
#define BRUSH_NAME_CHUNK	0x0110
#define BRUSH_STR_CHUNK	0x0120
#define BRUSH_RADIUS_CHUNK	0x0130
#define BRUSH_POINTCOUNT_CHUNK	0x0140
#define BRUSH_POINTDATA_CHUNK	0x0150
#define BRUSH_CLASSID_CHUNK	0x0160


class BrushClass
	{
public:
	int inkType;
	TSTR name;
	float str, radius;
	Tab<CurvePoint> points;
	Class_ID brushID;

	IOResult Load(ILoad *iload)
		{
		ULONG nb;
		IOResult res = IO_OK;

		int ct = 0;
	
	
		while (IO_OK==(res=iload->OpenChunk())) {
			switch (iload->CurChunkID()) {
				case BRUSH_CLASSID_CHUNK: 
					{
					res = iload->Read(&brushID,sizeof(Class_ID),&nb);
					break;
					}

				case BRUSH_INK_CHUNK: 
					{
					res = iload->Read(&inkType,sizeof(inkType),&nb);
					break;
					}
				case BRUSH_NAME_CHUNK: 
					{
					TCHAR n[200];
					int l;
					res = iload->Read(&l,sizeof(l),&nb);
					res = iload->Read(n,sizeof(TCHAR)*l,&nb);
					name.printf("%s",n);
					break;
					}
				case BRUSH_STR_CHUNK: 
					{
					res = iload->Read(&str,sizeof(str),&nb);
					break;
					}
				case BRUSH_RADIUS_CHUNK: 
					{
					res = iload->Read(&radius,sizeof(radius),&nb);
					break;
					}
				case BRUSH_POINTCOUNT_CHUNK: 
					{
					res = iload->Read(&ct,sizeof(ct),&nb);
					break;
					}
				case BRUSH_POINTDATA_CHUNK: 
					{
					CurvePoint pt;	
					res = iload->Read(&pt,sizeof(pt),&nb);
					points.Append(1,&pt);
					break;
					}



				}
			iload->CloseChunk();
			if (res!=IO_OK)  return res;
			}	

		return IO_OK;
		}
	IOResult Save(ISave *isave)
		{
		ULONG nb;

		isave->BeginChunk(BRUSH_CHUNK);
		isave->BeginChunk(BRUSH_INK_CHUNK);
		isave->Write(&inkType,sizeof(inkType),&nb);
		isave->EndChunk();	

		TCHAR *n = name.data();
		int l = name.Length()+1;


		isave->BeginChunk(BRUSH_NAME_CHUNK);
		isave->Write(&l,sizeof(l),&nb);
		isave->Write(n,sizeof(TCHAR)*l,&nb);
		isave->EndChunk();	

		isave->BeginChunk(BRUSH_STR_CHUNK);
		isave->Write(&str,sizeof(str),&nb);
		isave->EndChunk();	

		isave->BeginChunk(BRUSH_RADIUS_CHUNK);
		isave->Write(&radius,sizeof(radius),&nb);
		isave->EndChunk();	

		int ct = points.Count();

		isave->BeginChunk(BRUSH_POINTCOUNT_CHUNK);
		isave->Write(&ct,sizeof(ct),&nb);
		isave->EndChunk();	


		for (int i =0; i < ct; i++)
			{
			isave->BeginChunk(BRUSH_POINTDATA_CHUNK);
			CurvePoint pt = points[i];
			isave->Write(&pt,sizeof(pt),&nb);
			isave->EndChunk();	

			}

		isave->BeginChunk(BRUSH_CLASSID_CHUNK);
		isave->Write(&brushID,sizeof(Class_ID),&nb);
		isave->EndChunk();	

		isave->EndChunk();	
		return IO_OK;

		}

	};

typedef BOOL (CALLBACK* WTPACKET)(HCTX, int, LPVOID);
typedef HCTX (CALLBACK* WTOPEN)(HWND, LPLOGCONTEXTA, BOOL);
typedef UINT (CALLBACK* WTINFO) (UINT, UINT, LPVOID);
 
class PaintMod : public Modifier, public IPaintMod {	
	public:

//custom ink plugins
		ULONG customInkFlags;
		IInkClass *currentInk;

		Tab<BrushClass*> brushList;
		void FilloutBrushList();
		void AddBrush();
		void AddBrush(TSTR &name);
		void SetBrush();
		void RemoveBrush();

		ICurve *pCurve;


		ICurveCtl *curveCtl;

		HINSTANCE hWinTabDLL;               
		void LoadWinTabDLL();
		void UnLoadWinTabDLL();

		WTPACKET PaintWTPacket;
		WTOPEN PaintWTOpen;
		WTINFO PaintWTInfo;

#ifdef USETABLET
//tablet stuff
		UINT pressure;
#endif

		BOOL paintStroke;
		Tab<Point3> strokeList;
		HWND	viewhWnd;
		Matrix3 viewTM;
		int strokeState;
		float fpressure;
		int lag;
		float mapSize;
		int quadTree;
		BOOL showAffected;

		BOOL affectSize, affectStr;
		BOOL alignVec;
		
		BOOL useMirror;
		float offset;
		int mirrorDir;
		float mirrorDispSize;

		

		int hitTestType;



		IParamBlock2 *pblock;
		static PaintMode *paintMode;

		Tab<Point3> DeltaListOld;
		Tab<Point3> TempListOld;
		Tab<float> SelectVertListOld;

		int SelectMode;

		BYTE sel[1];

		static IObjParam *ip;
		
		PaintMod();
		~PaintMod() {
				DeleteAllRefsFromMe();
				pblock = NULL;
				}

static	float strength, eradius;
		float strengthsquared,eradiussquared;
static	float hardness;
static	float effect;
		int Addative;
		IPoint2 pp;

		BOOL useTablet;
		Color ringColor, normalColor, pressureColor;
		BOOL drawRing,drawNormal, drawPressure;
		BOOL updateOnMouseUp;


		HWND hWnd;
		HWND inkHWND;
static		int inkType;


		// From Animatable
		void DeleteThis() { delete this; }
		void GetClassName(TSTR& s) {s = GetString(IDS_PAINTMOD);}  
		virtual Class_ID ClassID() { return Class_ID(PAINTMOD_CLASSID);}		
		void BeginEditParams(IObjParam  *ip, ULONG flags,Animatable *prev);
		void EndEditParams(IObjParam *ip,ULONG flags,Animatable *next);		
		RefTargetHandle Clone(RemapDir& remap = NoRemap());
		TCHAR *GetObjectName() {return GetString(IDS_PAINTMOD);}

		// From modifier 
		ChannelMask ChannelsUsed()  {return PART_GEOM/*|PART_TOPO|PART_SELECT|PART_SUBSEL_TYPE*/;}
		ChannelMask ChannelsChanged() {return PART_GEOM|PART_SELECT|PART_SUBSEL_TYPE;}
		Class_ID InputType() {return defObjectClassID;}
		void ModifyObject(TimeValue t, ModContext &mc, ObjectState *os, INode *node);
		Interval LocalValidity(TimeValue t);

		// From BaseObject
		CreateMouseCallBack* GetCreateMouseCallBack() {return NULL;} 

		int NumRefs() {return 1;}
		RefTargetHandle GetReference(int i);
		void SetReference(int i, RefTargetHandle rtarg);

		int NumSubs() {return 1;}
		Animatable* SubAnim(int i){return pblock;}; 
		TSTR SubAnimName( int i  ) {return TSTR(GetString(IDS_PARAMETERS));};

		RefResult NotifyRefChanged( Interval changeInt,RefTargetHandle hTarget, 
		   PartID& partID, RefMessage message);


// JBW: direct ParamBlock access is added
		int	NumParamBlocks() { return 1; }					// return number of ParamBlocks in this instance
		IParamBlock2* GetParamBlock(int i) { if (i == 0) return pblock; 
											else return NULL;
												} // return i'th ParamBlock
		IParamBlock2* GetParamBlockByID(BlockID id) {if (pblock->ID() == id) return pblock ;
													else return  NULL; } // return id'd ParamBlock



		void EnterPaintAlign();
		void ExitPaintAlign();

		IOResult Load(ILoad *iload);
		IOResult Save(ISave *isave);

		int Display(TimeValue t, INode* inode, ViewExp *vpt, int flagst, ModContext *mc);
		void DrawCrossSection(Point3 a, Point3 align,Point3 alignDir, float length,float strenght, GraphicsWindow *gw);
		void DrawCrossSectionMirror(Point3 a, Point3 align,Point3 alignDir, float length,float strenght, GraphicsWindow *gw);


		void GetWorldBoundBox(TimeValue t,INode* inode, ViewExp *vpt, Box3& box, ModContext *mc);		

		BOOL inPaint;
		
		void ResetVerts();
		void ResetGraph();

		BOOL showVC;
		void ShadeVerts(Mesh *msh);

		INode* GetNodeFromModContext(ModContext *smd);
		PaintModData* GetPMDFromNode(INode *node);

		IOResult SaveLocalData(ISave *isave, LocalModData *pld);


		IOResult LoadLocalData(ILoad *iload, LocalModData **pld);

		void SetCustomInk();

//published functions		 
		//Function Publishing method (Mixin Interface)
		//******************************
		BaseInterface* GetInterface(Interface_ID id) 
			{ 
			if (id == PAINTMOD_INTERFACE) 
				return (IPaintMod*)this; 
			else 
				return FPMixinInterface::GetInterface(id);
			} 

		
		void SetBrush(int index);

		void	fnButtonPaint();
		void	fnButtonResetGraph();
		void	fnButtonAddBrush();
		void	fnButtonRemoveBrush();
		void	fnButtonReset();

		void	fnSetBrushByIndex(int index);
		void	fnSetBrushByName(TCHAR *name);

		int		fnCreateBrush(TCHAR *name);
		void	fnRemoveBrush(int index);
		int		fnNumberBrushes();

		void	fnSetBrushDataGraph(int index, Point2 p1, Point2 handle1, Point2 p2, Point2 handle2);
		void	fnSetBrushDataName(int index, TCHAR *name);
		void	fnSetBrushDataStrength(int index, float str);
		void	fnSetBrushDataRadius(int index, float r);
		void	fnSetBrushDataInk(int index, int ink);

		TCHAR*	fnGetBrushDataName(int index);
		float	fnGetBrushDataStrength(int index);
		float	fnGetBrushDataRadius(int index);
		int		fnGetBrushDataInk(int index);


	};


class AFRUndoList
	{
	public:
	int vertindex;
	Point3 delta;
	};

// Flags
#define CID_PAINT 	0x4f298c7a


class PaintMouseProc : public MouseCallBack {
	public:


		Tab<Matrix3> obtm,iobtm;

		Tab<PaintModData*> pmdList;
		Tab<INode*> nodeList;
		PaintMod *mod;
		IObjParam *ip;

		BOOL updateStack;
		

		void SetPoints();
		BOOL ApplyInk(Point3 pt, Point3 nnorm,Point3 snorm,  Point3 bary, DWORD faceIndex,PaintModData *pmd);


		PaintMouseProc(PaintMod *m,IObjParam *i) {mod=m;ip=i;}
		int proc(HWND hWnd, int msg, int point, int flags, IPoint2 m);
		void PaintMap(HWND hWnd,IPoint2 m);
		void HitMap(HWND hWnd,IPoint2 m);

		void MousePointDown(int point, int flags, IPoint2 m);
		void MousePointUp(int point, int flags, IPoint2 m);
		void MouseMove();
	};

class PaintMode : public CommandMode {
	public:
		ChangeFGObject fgProc;
		PaintMouseProc proc;
		IObjParam *ip;
		PaintMod *mod;

		PaintMode(PaintMod *m,IObjParam *i) 
			: fgProc(m), proc(m,i) {ip=i;mod=m;}

		int Class() {return MOVE_COMMAND;}		
		int ID() {return CID_PAINT;}
		MouseCallBack *MouseProc(int *numPoints) {*numPoints=2;return &proc;}
		ChangeForegroundCallback *ChangeFGProc() {return &fgProc;}
		BOOL ChangeFG(CommandMode *oldMode) {return oldMode->ChangeFGProc() != &fgProc;}
		void EnterMode();
		void ExitMode();
	};

// A restore object to save the position of control points.

class MoveRestore : public RestoreObj {
	public:		
		PaintMod *mod;
		PaintModData *pmd;
		Tab<AFRUndoList> PListUndo;
		MoveRestore(PaintMod *mod, PaintModData *m) {
			this->mod = mod;
			pmd = m;
			PListUndo.ZeroCount();
			for (int i=0; i<pmd->TempList.Count(); i++) {
				if (Length(pmd->TempList[i]) > 0.000f)
					{
					AFRUndoList t;
					t.vertindex = i;
					t.delta = pmd->TempList[i];
					PListUndo.Append(1,&t,1);
//					PListRedo.Append(1,&t,1);
					}
				}
			}

		void Restore(int isUndo) {
			// if we're undoing, save a redo state

			if (isUndo) {
				for (int i=0; i<PListUndo.Count(); i++) {
					pmd->DeltaList[PListUndo[i].vertindex] -= PListUndo[i].delta;
//					mod->GetPt(i) = undo[i];
					}
				pmd->redoNormalList=TRUE;
				mod->NotifyDependents(FOREVER,PART_GEOM,REFMSG_CHANGE);
				}
			}

		void Redo() {
			for (int i=0; i<PListUndo.Count(); i++) {
				pmd->DeltaList[PListUndo[i].vertindex] += PListUndo[i].delta;
				}
			pmd->redoNormalList=TRUE;
			mod->NotifyDependents(FOREVER,PART_GEOM,REFMSG_CHANGE);
			}
	TSTR Description()

		{ return TSTR(_T("Paint Deform")); }
	};



class PaintClassDesc:public ClassDesc2 {
	public:
	int 			IsPublic() { return 1; }
	void *			Create(BOOL loading = FALSE);
	const TCHAR *	ClassName() { return GetString(IDS_RB_PMOD); }
	SClass_ID		SuperClassID() { return OSM_CLASS_ID; }
	Class_ID		ClassID() { return Class_ID(PAINTMOD_CLASSID); }
	const TCHAR* 	Category() { return GetString(IDS_CHARITYWARE);}

	const TCHAR*	InternalName() { return _T("PaintMod"); }	// returns fixed parsable name (scripter-visible name)
	HINSTANCE		HInstance() { return hInstance; }			// returns owning module handle

	};

class PaintDeformer : public Deformer {
	public:		

		PaintMod *mod;
		PaintModData *pmd;
		PaintDeformer(PaintMod *m,PaintModData *md) 
				{
				mod = m;
				pmd = md;
				}
		Point3 Map(int i, Point3 p);
	};


class MyEnumProc : public DependentEnumProc 
	{
      public :
      virtual int proc(ReferenceMaker *rmaker); 
      INodeTab Nodes;              
	};



#endif

