

#include "mods.h"

static int modCount = 0;


BOOL PaintMouseProc::ApplyInk(Point3 pt, Point3 nnorm, Point3 snorm, Point3 bary, DWORD faceIndex, PaintModData *pmd)
	{
	updateStack = TRUE;
	if (mod->inkType >(NUMBER_DEFAULTINKS-1))
		{
		if (mod->currentInk)
			return mod->currentInk->ApplyInk(pt,  snorm, bary, faceIndex,
							&pmd->msh,
							mod->strength, mod->eradius,mod->fpressure, 
							mod->pCurve,
							pmd->oPointList, pmd->DeltaList, pmd->TempList,
							pmd->tmToWorld,
							mod->Addative,
							updateStack);
		}

	BOOL changed = FALSE;
	float minx,maxx,miny,maxy,minz,maxz;
	minx = pt.x - mod->eradius; 
	miny = pt.y - mod->eradius;
	minz = pt.z - mod->eradius;

	maxx = pt.x + mod->eradius; 
	maxy = pt.y + mod->eradius;
	maxz = pt.z + mod->eradius;

	Tab<Point3> holdPoints;
	if (mod->inkType==INK_SMOOTH)
		holdPoints = pmd->TempList;
	for (int vc = 0; vc < pmd->oPointList.Count();vc++)
		{
		if ( (mod->inkType == INK_ORIGINALNORMAL) && (pmd->objectType == PAINTMESH_OBJECT))
			nnorm = pmd->origVNorms[vc].GetNormal(0xffff)*mod->strength ;
//look bounding box
		Point3 p = pmd->oPointList[vc];
	 	if ( (p.x >= minx)  && (p.x <= maxx) &&
			 (p.y >= miny)  && (p.y <= maxy) && 
			 (p.z >= minz)  && (p.z <= maxz) ) 
			{
			float d = Length(pt-p);
			float eradius = mod->eradius;
			if ((mod->useTablet) && (mod->affectSize))
				eradius = eradius * mod->fpressure;
			if (d < eradius )
				{
				changed = TRUE;
				if (mod->inkType==INK_SMOOTH)
					{
					if (pmd->objectType == PAINTMESH_OBJECT)
						{
						float f;
						f = d/eradius;
						f = mod->pCurve->GetValue(0, f, FOREVER, TRUE);
						
//if addative check distance
						Point3 vec;
						f = (f*mod->strength * mod->fpressure);
						Point3 ovec = pmd->meshSearchTree.SmoothVert(vc,mod->Addative,&holdPoints);
						vec = ovec;
						
						pmd->TempList[vc] += vec * f;
						float currentLen = Length(ovec);
						float newLen = Length(pmd->TempList[vc]);
						if ( newLen > currentLen)
							{
							pmd->TempList[vc] = ovec;
							}
						}
					
					}
				else if (mod->inkType==INK_REVERT)
					{
					if (vc < pmd->DeltaList.Count())
						{
						float f;
						f = d/eradius;
						f = mod->pCurve->GetValue(0, f, FOREVER, TRUE);
						
//if addative check distance
						Point3 vec;
						Point3 norm = pmd->DeltaList[vc];
						f = (f*mod->strength * mod->fpressure);
						if (f < 0.0f) f = 0.0f;
						if (f > 1.0f) f = 1.0f;
						
						vec = pmd->TempList[vc] + pmd->DeltaList[vc];
						if (mod->SelectMode & MOUSE_ALT)
							vec = -vec * f;
						else vec = vec * f;
						pmd->TempList[vc] += vec;


						}
					}
//select ink
				else if (mod->inkType ==INK_SELECTVERTS)
					{
					if (pmd->objectType == PAINTMESH_OBJECT)
						{

						float f;
						f = d/mod->eradius;
						f = mod->pCurve->GetValue(0, f, FOREVER, TRUE);
						
//if addative check distance
						f = f * mod->fpressure * mod->strength;
						if (mod->SelectMode & MOUSE_ALT)
							{
							pmd->SelectVertList[vc] -= f;
							changed = TRUE;
							}
						else 
							{
							if (mod->Addative)
								pmd->SelectVertList[vc] += f;
							else 
								{
								if (f > pmd->SelectVertList[vc])
									pmd->SelectVertList[vc] = f;
								}
							changed = TRUE;
							}
						if (pmd->SelectVertList[vc] > 1.0f) pmd->SelectVertList[vc] = 1.0f;
						if (pmd->SelectVertList[vc] < -1.0f) pmd->SelectVertList[vc] = -1.0f;
						}

							

					}
				else if (mod->inkType ==INK_SELECTFACES)
					{
					if (pmd->objectType == PAINTMESH_OBJECT)
						{
						if (mod->SelectMode & MOUSE_ALT)
							{
							pmd->meshSearchTree.SelectFace(vc, pmd->selectedFaceList,FALSE);
							}
						else 
							{
							pmd->meshSearchTree.SelectFace(vc, pmd->selectedFaceList,TRUE);
							}
						changed = TRUE;

						}
							

					}
				else if (mod->inkType!=INK_SELECTVERTS)
					{
					if (vc < pmd->DeltaList.Count())
						{
						float f;
						f = d/eradius;
						f = mod->pCurve->GetValue(0, f, FOREVER, TRUE);
						
//if addative check distance
						Point3 vec;
						if (mod->affectStr)
							vec = (nnorm * f) * mod->fpressure ;
						if (mod->SelectMode & MOUSE_ALT)
							vec = -vec;
								
						if (!mod->Addative)
							{
							float l = Length(vec);
							if (l > Length(pmd->TempList[vc]) )
								pmd->TempList[vc] = vec;
	
							}	
						else 
							{
							pmd->TempList[vc] += vec;
							}


						changed = TRUE;
						}
					}

				}

			}

		}

	return changed;
	}

void PaintMouseProc::PaintMap(HWND hWnd,IPoint2 m)
	{

	ViewExp *vpt = ip->GetActiveViewport();

	if (!vpt) return;

	Ray ray, wray;
	TimeValue t = ip->GetTime();	
	Point3 norm, pt;
	Interval valid;
	BOOL changed = FALSE;	


	// Calculate a ray from the mouse point
	int dist;
	IPoint2 inc; 

	dist = Length(m-mod->pp)/mod->lag;
	if (dist == 0) dist = 1;
	inc = (m-mod->pp)/dist;

	

	float fdist;
	Point2 finc,initM;
	fdist = Length(m-mod->pp)/(float) mod->lag;
	Point2 mt,ppt;
	mt.x = m.x;
	mt.y = m.y;
	ppt.x = mod->pp.x;
	ppt.y = mod->pp.y;
	finc = (mt-ppt)/fdist;


	IPoint2 tp = mod->pp;
	mod->pp = m;

	m = tp;
	initM.x = m.x;
	initM.y = m.y;


	Tab<Point3> xynorm,znorm,xynormlocal;
	Tab<Ray> rray;
	float pirad;

	pirad = 3.14f/mod->eradius;


	for (int i=0; i<nodeList.Count(); i++) 
		{

		Matrix3 tobtm  = obtm[i];
		Matrix3 tiobtm = iobtm[i];


//xynorm
		Matrix3 TM;
		vpt->GetAffineTM(TM);
		Matrix3 invTM = Inverse(TM);
		Point3 txynorm;
		txynorm.x = (float) inc.x;
		txynorm.y = -(float) inc.y;
		txynorm.z = 0.0f;
		txynorm = VectorTransform(invTM, txynorm);
		txynorm = VectorTransform(tiobtm,txynorm);
		txynorm = Normalize(txynorm)*mod->strength;
		xynorm.Append(1,&txynorm,1);
//move z
		Ray tempray;
		vpt->MapScreenToWorldRay(float(m.x), float(m.y),tempray);

		Ray tempray2;
		tempray2.dir = VectorTransform(tiobtm, tempray.dir);
		rray.Append(1,&tempray2,1);

		Point3 tznorm;
		tznorm = VectorTransform(tiobtm,tempray.dir);
		tznorm = -Normalize(tznorm)*mod->strength;
		znorm.Append(1,&tznorm,1);

		}




	for (int j = 0;j < dist;j++)
		{
		vpt->MapScreenToWorldRay(float(m.x), float(m.y),wray);
	
		float fz = 1.0f;
		BOOL first = TRUE;
		BOOL hit = FALSE;

		Point3 fp,fnorm;
		int hitNode;
		DWORD faceIndex;
		Point3 bary;

		for (int i=0; i<nodeList.Count(); i++) 
			{


	// Get the object from the node
			INode *node = nodeList[i];


			ray.p   = iobtm[i] * wray.p;
			ray.dir = rray[i].dir;
	
		// See if we hit the object
			Point3 bry;
		

			DWORD faceIndexTemp;

			BOOL tempHit = FALSE;


			Point3 p;
			float z;
			
			

			if (mod->hitTestType == HITTEST_QUADTREE) 
				{
				IPoint2 m2 = m;
				tempHit = pmdList[i]->meshSearchTree.HitQuadTree(m2,faceIndexTemp, p,bry,z);
				}

			if (tempHit) 
				{

				if ((first) || (z>fz ))
					{
					bary = bry;
					fz = z;
					first = FALSE;
					hit = TRUE;
					fnorm = pmdList[i]->fnorms[faceIndexTemp];
					fp = p;
					hitNode = i;
					pmdList[i]->paintPoint = fp;
					pmdList[i]->paintNorm = fnorm;
					faceIndex  = faceIndexTemp;

					}
				}
			}


		if (hit)
			{
			mod->strokeList.Append(1,&fp);

			Point3 wPoint, wNorm;
			wPoint = pmdList[hitNode]->paintPoint * obtm[hitNode];
			wNorm = VectorTransform(pmdList[hitNode]->paintNorm, obtm[hitNode]);

			for (i=0; i<nodeList.Count(); i++) 
				{
				Point3 lPoint, lNorm;
				if (i!=hitNode)
					{
					

					lPoint = wPoint * iobtm[i];
					lNorm = VectorTransform(wNorm, iobtm[i]);

					pmdList[i]->paintPoint = lPoint;
					pmdList[i]->paintNorm = lNorm;
					}

				pmdList[i]->tmToWorld =  obtm[hitNode];
				Point3 snorm = pmdList[i]->paintNorm;

				Point3 nnorm(0.0f,0.0f,0.0f);
				Point3 pt = pmdList[i]->paintPoint;
//move normal
				if (mod->inkType == INK_NORMAL)
					nnorm = Normalize(pmdList[i]->paintNorm)*mod->strength ;
				else if ( (mod->inkType == INK_ORIGINALNORMAL) && (pmdList[hitNode]->objectType == PAINTMESH_OBJECT))
					nnorm = pmdList[hitNode]->origFNorms[faceIndex]*mod->strength ;
//move xy
				else if (mod->inkType == INK_XYSCREEN)
					{
					nnorm = xynorm[i] ;
					}
//move z
				else if (mod->inkType == INK_ZSCREEN)
					{
					nnorm = znorm[i] ;

					}
//move x local
				else if (mod->inkType == INK_XYLOCAL)
					{
					nnorm.x = mod->strength  ;
					nnorm.y = 0.0f;
					nnorm.z = 0.0f;

					}
//move y local
				else if (mod->inkType == INK_XYLOCAL)
					{
					nnorm.y = mod->strength ;
					nnorm.x = 0.0f;
					nnorm.z = 0.0f;

					}
//move z local
				else if (mod->inkType == INK_ZLOCAL)
					{
					nnorm.z = mod->strength ;
					nnorm.x = 0.0f;
					nnorm.y = 0.0f;

					}
//smooth
				else if (mod->inkType == INK_SMOOTH)
					{
					nnorm.z = 0.0f;
					nnorm.x = 0.0f;
					nnorm.y = 0.0f;

					}
//tesselate
				else if (mod->inkType == 6)
					{

					}
//select
				else if (mod->inkType == INK_SELECTVERTS)
					{

					}

				pmdList[i]->alignDir = Normalize(nnorm);

				if (ApplyInk(pt, nnorm,snorm, bary, faceIndex, pmdList[i])) 
					changed = TRUE;
				if (mod->useMirror)
					{
					if (mod->mirrorDir == 0)
						{
						pt.x = -pt.x + mod->offset;
						nnorm.x = -nnorm.x;
						}
					else if (mod->mirrorDir == 1)
						{
						pt.y= -pt.y + mod->offset;
						nnorm.y = -nnorm.y;

						}
					else if (mod->mirrorDir == 2)
						{
						pt.z = -pt.z + mod->offset;
						nnorm.z = -nnorm.z;
						}
					if (ApplyInk(pt, nnorm, snorm, bary, faceIndex, pmdList[i])) changed = TRUE;
					}
				}
			}
		initM.x += finc.x;
		initM.y += finc.y;

		m.x = (int)initM.x;
		m.y = (int)initM.y;
		}
	if (changed)
		{
  
		if ( (!mod->updateOnMouseUp) && (updateStack))
			{
			if (mod->inkType==INK_SELECTVERTS)
				{
				mod->NotifyDependents(Interval(t,t), PART_SELECT, REFMSG_MOD_EVAL);
				mod->NotifyDependents(Interval(t,t), PART_SUBSEL_TYPE, REFMSG_MOD_EVAL);
				mod->NotifyDependents(Interval(t,t), PART_DISPLAY, REFMSG_MOD_EVAL);
				}
			else mod->NotifyDependents(Interval(t,t), PART_GEOM, REFMSG_MOD_EVAL);
			mod->ip->RedrawViews(t);
			}
		else
			{
			mod->ip->ForceCompleteRedraw(FALSE);
			}

		}
	
	ip->ReleaseViewport(vpt);

	}


void PaintMouseProc::HitMap(HWND hWnd,IPoint2 m)
	{

	ViewExp *vpt = ip->GetActiveViewport();
	if (!vpt) 
		{
		return;
		}

	TimeValue t = ip->GetTime();	
	
	Matrix3 viewTm; 
	vpt->GetAffineTM(viewTm);

	if (!(mod->viewTM == viewTm))
		{	
		for (int i = 0; i < pmdList.Count(); i++)
			pmdList[i]->redoNormalList = TRUE;
		mod->NotifyDependents(Interval(t,t), PART_DISPLAY, REFMSG_MOD_EVAL);
		}

	Ray ray, wray;
	Point3 norm, pt;
	Interval valid;
	BOOL changed = FALSE;	


	// Calculate a ray from the mouse point

	mod->pp = m;

	IPoint2 inc; 
	int dist;

	dist = Length(m-mod->pp)/mod->lag;
	if (dist == 0) dist = 1;
	inc = (m-mod->pp)/dist;


	Tab<Ray> rray;

	vpt->MapScreenToWorldRay(float(m.x), float(m.y),wray);
	
	Tab<Point3> xynorm,znorm,xynormlocal;
	
	for (int i=0; i<nodeList.Count(); i++) 
		{
		if (!pmdList[i]->isMesh)
			{
			Ray	tempray;
			vpt->MapScreenToWorldRay(float(m.x), float(m.y),tempray);

			Ray tempray2;
			tempray2.dir = VectorTransform(iobtm[i], tempray.dir);
			rray.Append(1,&tempray2,1);
			}
		else
			{
			Ray tempray2;
			rray.Append(1,&tempray2,1);
			}

//xynorm
		Matrix3 tiobtm = iobtm[i];
		Matrix3 TM;
		vpt->GetAffineTM(TM);
		Matrix3 invTM = Inverse(TM);
		Point3 txynorm;
		txynorm.x = (float) inc.x;
		txynorm.y = -(float) inc.y;
		txynorm.z = 0.0f;
		txynorm = VectorTransform(invTM, txynorm);
		txynorm = VectorTransform(tiobtm,txynorm);
		txynorm = Normalize(txynorm)*mod->strength;
		xynorm.Append(1,&txynorm,1);
//move z
		Ray tempray;
		vpt->MapScreenToWorldRay(float(m.x), float(m.y),tempray);

		Ray tempray2;
		tempray2.dir = VectorTransform(tiobtm, tempray.dir);
		rray.Append(1,&tempray2,1);

		Point3 tznorm;
		tznorm = VectorTransform(tiobtm,tempray.dir);
		tznorm = -Normalize(tznorm)*mod->strength;
		znorm.Append(1,&tznorm,1);

		}	




	BOOL hit = FALSE;
	int hitNode = 0;
	DWORD faceIndex;
	Point3 bary;
	float z = -1.0f;
	BOOL first = TRUE;

	for (i=0; i<nodeList.Count(); i++) 
		{
		float tz;
		Point3 fp;
		Point3 norm;
		BOOL tempHit = FALSE;
		if (pmdList[i]->isMesh)
			{
			IPoint2 m2 = m;
			;
			if (mod->hitTestType == HITTEST_QUADTREE)
				{
				Point3 p;
				tempHit = pmdList[i]->meshSearchTree.HitQuadTree(m2,faceIndex, p, bary,tz);
				if (tempHit)
					{
					fp = p;
					norm = pmdList[i]->fnorms[faceIndex];

					}	
				Point3 nnorm(0.0f,0.0f,0.0f);
				Point3 pt = pmdList[i]->paintPoint;
//move normal
				if (mod->inkType == INK_NORMAL)
					nnorm = Normalize(pmdList[i]->paintNorm)*mod->strength ;
				else if ( (tempHit) && (mod->inkType == INK_ORIGINALNORMAL) && (pmdList[i]->objectType == PAINTMESH_OBJECT))
					nnorm = pmdList[i]->origFNorms[faceIndex]*mod->strength ;
//move xy

//move x local
				else if (mod->inkType == INK_XYLOCAL)
					{
					nnorm.x = mod->strength  ;
					nnorm.y = 0.0f;
					nnorm.z = 0.0f;

					}
//move y local
				else if (mod->inkType == INK_XYLOCAL)
					{
					nnorm.y = mod->strength ;
					nnorm.x = 0.0f;
					nnorm.z = 0.0f;

					}
//move z local
				else if (mod->inkType == INK_ZLOCAL)
					{
					nnorm.z = mod->strength ;
					nnorm.x = 0.0f;
					nnorm.y = 0.0f;

					}
//smooth
				else if (mod->inkType == INK_SMOOTH)
					{
					nnorm.z = 0.0f;
					nnorm.x = 0.0f;
					nnorm.y = 0.0f;

					}
//tesselate
				else if (mod->inkType == 6)
					{

					}
//select
				else if (mod->inkType == INK_SELECTVERTS)
					{

					}

				pmdList[i]->alignDir = Normalize(nnorm);
				
				}
			}

		pmdList[i]->draw = FALSE;
		if ((tempHit) && ((tz>z) || first))
			{
			
			hit = TRUE;
			hitNode = i;
			pmdList[i]->paintNorm = norm;
			pmdList[i]->paintPoint = fp;
			z = tz;
			first = FALSE;
			}
		

		}

	Point3 wPoint, wNorm;
	wPoint = pmdList[hitNode]->paintPoint * obtm[hitNode];
	wNorm = VectorTransform(pmdList[hitNode]->paintNorm, obtm[hitNode]);

	for (i=0; i<nodeList.Count(); i++) 
		{
		if (i!=hitNode)
			{
			Point3 lPoint, lNorm;

			lPoint = wPoint * iobtm[i];
			lNorm = VectorTransform(wNorm, iobtm[i]);

			pmdList[i]->paintPoint = lPoint;
			pmdList[i]->paintNorm = lNorm;

			}		

		}

//transform point and norm to world space
//transform the others into their space

	if (hit)
		{
		pmdList[hitNode]->draw = TRUE;
		mod->ip->ForceCompleteRedraw(FALSE);
		}

	ip->ReleaseViewport(vpt);

}

void PaintMouseProc::SetPoints()
	{


	TimeValue t = ip->GetTime();	

	for (int i =0;  i < nodeList.Count(); i++)
		{
		INode *node = nodeList[i];

		PaintModData* pmd = mod->GetPMDFromNode(nodeList[i]);
		pmd->setPoints = TRUE;
		}
	}

void PaintMouseProc::MousePointDown(int point, int flags, IPoint2 m)
	{

	if (mod->inkType > (NUMBER_DEFAULTINKS-1))
		mod->currentInk->StartStroke();

	if (!mod->useTablet) mod->fpressure = 1.0f;
	mod->strokeList.ZeroCount();

	modCount = 0;
	Point3 p(0.0f,0.0f,0.0f);
	mod->pp = m;
	for (int j =0; j < pmdList.Count(); j++)
		{
		for (int i = 0; i < pmdList[j]->DeltaList.Count(); i++)
			{
			pmdList[j]->TempList[i] = p;
			}
		pmdList[j]->redoNormalList = FALSE;
		}


	mod->SelectMode = flags;
	if ((!(flags & MOUSE_ALT)) && (!(flags & MOUSE_CTRL)))
		{
		if (mod->inkType == INK_SELECTFACES)
			{
			for (int j =0; j < pmdList.Count(); j++)
				pmdList[j]->selectedFaceList.ClearAll();
			}
		if (mod->inkType == INK_SELECTVERTS)
			{
			for (int j =0; j < pmdList.Count(); j++)
				{

				for (int i = 0; i < pmdList[j]->SelectVertList.Count(); i++)
					{
					pmdList[j]->SelectVertList[i] = FALSE;
					}
				}
			}

		}
	ip->RedrawViews(ip->GetTime(),REDRAW_BEGIN);
	mod->strokeState = INSTROKE;
	
	}

void PaintMouseProc::MousePointUp(int point, int flags, IPoint2 m)
	{
	SetPoints();

	if (mod->inkType > (NUMBER_DEFAULTINKS-1))
		mod->currentInk->EndStroke();

	mod->strokeList.ZeroCount();
	Point3 p(0.0f,0.0f,0.0f);
	if (!theHold.Holding() ) 
		{
		theHold.Begin();
		}
	for (int j =0; j < pmdList.Count(); j++)
		theHold.Put(new MoveRestore(mod, pmdList[j]));
	theHold.Accept(TSTR(_T("Paint Deform")));
				

	for (j =0; j < pmdList.Count(); j++)
		{
		for (int i = 0; i < pmdList[j]->DeltaList.Count(); i++)
			{
			pmdList[j]->DeltaList[i] += pmdList[j]->TempList[i];
			pmdList[j]->TempList[i] = p;
			}

		pmdList[j]->redoNormalList = TRUE;
		}

	mod->NotifyDependents(Interval(ip->GetTime(),ip->GetTime()), PART_GEOM, REFMSG_MOD_EVAL);
	ip->RedrawViews(ip->GetTime(),REDRAW_END);

	
	}

#define PAINTMODE	0 
#define SIZEMODE	1
#define STRMODE		2


int PaintMouseProc::proc(
		HWND hWnd, int msg, int point, int flags, IPoint2 m)
	{
	static int mode;
	static IPoint2 prevPoint;
	static float str,size;
	int i;
	Point3 p(0.0f,0.0f,0.0f);
	static IPoint2 oldPoint;
	switch (msg) {
		case MOUSE_POINT:
			if (point==0) 
				{	
//grow 	size
				if ( (flags & MOUSE_SHIFT) && (flags & MOUSE_ALT) && (flags & MOUSE_LBUTTON) )
					{
					mode = SIZEMODE;
					prevPoint = m;
					mod->pblock->GetValue(paint_eradius,ip->GetTime(),size,FOREVER);

					}
				else if ( (flags & MOUSE_SHIFT) && (flags & MOUSE_CTRL) && (flags & MOUSE_LBUTTON) )
					{
					mode = STRMODE;
					prevPoint = m;
					mod->pblock->GetValue(paint_str,ip->GetTime(),str,FOREVER);
					}
				else 
					{
					mode = PAINTMODE;
					MousePointDown(point, flags, m);
					PaintMap(hWnd,m);
					}
				} 
			else 
				{
//add hold here
				if (mode == PAINTMODE)
					MousePointUp(point, flags, m);
				mod->strokeState = OUTOFSTROKE;
				}
			break;

		case MOUSE_MOVE: {
			if (mode == PAINTMODE)
				{
				PaintMap(hWnd,m);
				}
			else
				{
				
				if (mode == SIZEMODE)
					{
					float value = size + ((m.x - prevPoint.x)/5.0f );
					mod->pblock->SetValue(paint_eradius,ip->GetTime(),value);
					ip->RedrawViews(ip->GetTime(),REDRAW_NORMAL);

					}
				else if (mode == STRMODE)
					{
					float value = str+((m.x - prevPoint.x)/5.0f);
					mod->pblock->SetValue(paint_str,ip->GetTime(),value);
					ip->RedrawViews(ip->GetTime(),REDRAW_NORMAL);

					}
				}

			break;
			}

		case MOUSE_ABORT:
			{

			if (mod->inkType > (NUMBER_DEFAULTINKS-1))
				mod->currentInk->CancelStroke();

			mod->strokeState = OUTOFSTROKE;
			mod->strokeList.ZeroCount();
			for (int j =0; j < pmdList.Count(); j++)
				{

				for (i = 0; i < pmdList[j]->DeltaList.Count(); i++)
					{
					pmdList[j]->TempList[i] = p;
					}
				}
			mod->NotifyDependents(Interval(ip->GetTime(),ip->GetTime()), PART_GEOM, REFMSG_MOD_EVAL);

			ip->RedrawViews(ip->GetTime(),REDRAW_END);
			break;
			}

		case MOUSE_FREEMOVE:
			if (m!=oldPoint)
				{
				if ((!mod->useTablet) || (mod->strokeState == OUTOFSTROKE))
					{
					SetCursor(ip->GetSysCursor(SYSCUR_SELECT));
					HitMap(hWnd,m);
					}
				else if (mod->strokeState == STARTSTROKE) 
					{
					MousePointDown(0, flags, m);
					}
				else if (mod->strokeState == ENDSTROKE) 
					{
					MousePointUp(1, flags, m);
					}
				else if (mod->strokeState == INSTROKE) 
					{
					PaintMap(hWnd,m);
					}
				oldPoint = m;
				}
			break;
		}
	return TRUE;
	}

void PaintMode::EnterMode()
	{
	proc.pmdList.ZeroCount();

	ModContextList mcList;
	INodeTab nodeList;

	proc.nodeList.ZeroCount();

	ip->GetModContexts(mcList,nodeList);
	proc.obtm.ZeroCount();
	proc.iobtm.ZeroCount();
	TimeValue t = ip->GetTime();
	
	for (int i = 0; i < nodeList.Count(); i++)
		{
//get node list
		PaintModData* pmd = mod->GetPMDFromNode(nodeList[i]);
		if (pmd)
			{
			proc.pmdList.Append(1,&pmd);
			proc.nodeList.Append(1,&nodeList[i]);

			Matrix3 tobtm  = nodeList[i]->GetObjectTM(t);
			Matrix3 tiobtm = Inverse(tobtm);

			proc.obtm.Append(1,&tobtm,1);
			proc.iobtm.Append(1,&tiobtm,1);
			pmd->redoNormalList = TRUE;

			}

		}


	nodeList.DisposeTemporary();

	mod->EnterPaintAlign();	
	mod->inPaint = TRUE;
	
	}

void PaintMode::ExitMode()
	{
	mod->ExitPaintAlign();	
	mod->inPaint = FALSE;
	}
